# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Backup Handler
"""

import logging
import os
import shutil
import socket
import subprocess
import sys
import warnings

from wuttjamaican.util import parse_bool, parse_list

from rattail.app import GenericHandler


log = logging.getLogger(__name__)


class BackupHandler(GenericHandler):
    """
    Base class and default implementation for backup handler.
    """

    def execute(self, params):
        """
        Primary method for callers to invoke.  Runs a backup according
        to params and/or config.
        """
        if params['list_dbs']:

            mysql_names = self.get_mysql_names()
            if mysql_names:
                width = max([len(name) for name in mysql_names])
                sys.stdout.write("\nmysql\n")
                sys.stdout.write("{}\n".format('-' * width))
                for name in mysql_names:
                    sys.stdout.write("{}\n".format(name))

            postgres_names = self.get_postgres_names()
            if postgres_names:
                width = max([len(name) for name in postgres_names])
                sys.stdout.write("\npostgres\n")
                sys.stdout.write("{}\n".format('-' * width))
                for name in postgres_names:
                    sys.stdout.write("{}\n".format(name))

            if mysql_names or postgres_names:
                sys.stdout.write("\n")
            else:
                sys.stdout.write("no databases found\n")

        else: # not listing dbs, so will do dbdump and/or rsync and/or borg

            if params['dbdump'] and params['no_dbdump']:
                raise RuntimeError("Must specify either --dbdump or --no-dbdump, but not both")
            if params['dump_tables'] and params['no_dump_tables']:
                raise RuntimeError("May specify either --dump-tables or --no-dump-tables, but not both")
            if params['rsync'] and params['no_rsync']:
                raise RuntimeError("May specify either --rsync or --no-rsync, but not both")
            if params['borg'] and params['no_borg']:
                raise RuntimeError("May specify either --borg or --no-borg, but not both")

            # we dump dbs by default, unless user or config file says not to
            dbdump = params['dbdump']
            if not dbdump:
                if params['no_dbdump']:
                    dbdump = False
                else:
                    dbdump = self.config.getbool('rattail.backup', 'dbdump.enabled', usedb=False)
                    if dbdump is None:
                        dbdump = self.config.getbool('rattail.backup', 'dbdump',
                                                     usedb=False, ignore_ambiguous=True)
                        if dbdump is not None:
                            warnings.warn(f"URGENT: instead of 'dbdump', "
                                          f"you should set 'dbdump.enabled'",
                                          DeprecationWarning, stacklevel=2)
                        else:
                            dbdump = True
            if dbdump:

                outdir = params['dbdump_output']
                if not outdir:
                    outdir = self.config.get('rattail.backup', 'dbdump.output', default='/root/data')
                if not os.path.exists(outdir):
                    os.mkdir(outdir)

                if params['included_dbs']:
                    include = params['included_dbs']
                else:
                    include = self.config.getlist('rattail.backup', 'dbdump.include', usedb=False)

                if params['excluded_dbs']:
                    exclude = params['excluded_dbs']
                else:
                    exclude = self.config.getlist('rattail.backup', 'dbdump.exclude', usedb=False)

                if params['dump_tables']:
                    tables = True
                elif params['no_dump_tables']:
                    tables = False
                else:
                    tables = self.config.getbool('rattail.backups', 'dbdump.dump_tables', usedb=False, default=False)

                self.dump_mysql_databases(outdir, include=include, exclude=exclude, tables=tables,
                                          verbose=params['verbose'], dry_run=params['dry_run'])
                self.dump_postgres_databases(outdir, include=include, exclude=exclude, tables=tables,
                                             verbose=params['verbose'], dry_run=params['dry_run'])

            # we do *not* run rsync by default, unless user or config file says to
            rsync = params['rsync']
            rsync_last = False
            if not rsync:
                if params['no_rsync']:
                    rsync = False
                else:
                    rsync = self.config.getbool('rattail.backup', 'rsync.enabled', usedb=False)
                    if rsync is None:
                        rsync = self.config.getbool('rattail.backup', 'rsync',
                                                    usedb=False, ignore_ambiguous=True)
                        if rsync is not None:
                            warnings.warn(f"URGENT: instead of 'rsync', "
                                          f"you should set 'rsync.enabled'",
                                          DeprecationWarning, stacklevel=2)
                        else:
                            rsync = False
            if rsync:
                if self.config.getbool('rattail.backup', 'rsync.last', usedb=False, default=False):
                    rsync_last = True

            # we do *not* run borg by default, unless user or config file says to
            borg = False
            if params['borg'] or (not params['no_borg'] and self.config.getbool('rattail.backup', 'borg.enabled', usedb=False, default=False)):
                borg = True

            # okay now run rsync and/or borg (in whichever order) depending on params
            if rsync and not rsync_last:
                self.run_rsync(params)
            if borg:
                self.run_borg(params)
            if rsync and rsync_last:
                self.run_rsync(params)

    def run_rsync(self, params):
        self.rsync_files(include=params['rsync_included_prefixes'] or None,
                         exclude=params['rsync_excluded_prefixes'] or None,
                         verbose=params['verbose'],
                         dry_run=params['dry_run'])

    def run_borg(self, params):
        remotes = (parse_list(params['borg_remotes']) or
                   self.config.getlist('rattail.backup', 'borg.remotes',
                                       usedb=False))
        self.make_borg_archives(remotes,
                                include=params['borg_included_prefixes'] or params['rsync_included_prefixes'] or None,
                                exclude=params['borg_excluded_prefixes'] or params['rsync_excluded_prefixes'] or None,
                                tag=params['borg_tag'],
                                verbose=params['verbose'],
                                dry_run=params['dry_run'])

    def get_path_includes(self, typ='rsync'):
        return self.config.getlist('rattail.backup', f'{typ}.include', usedb=False, default=[
            '/etc',
            '/home',
            '/opt',
            '/root',
            '/srv',
            '/usr/local',
            '/var',
        ])

    def get_default_excludes(self):
        return [
            '/var/cache/',
        ]

    def get_path_excludes(self, typ='rsync'):
        return self.config.getlist('rattail.backup', '{}.exclude'.format(typ), usedb=False)

    def validate_prefixes(self, prefixes):
        for prefix in prefixes:
            if not prefix.startswith('/'):
                raise RuntimeError("Prefix is not absolute path: {}".format(prefix))

        for prefix in prefixes:
            others = list(prefixes)
            others.remove(prefix)
            for other in others:
                if other.startswith(prefix):
                    raise RuntimeError("Prefix {} is redundant due to prefix {}".format(
                        other, prefix))

    def rsync_files(self, include=None, exclude=None, verbose=False, dry_run=False):
        remote_host = self.config.require('rattail.backup', 'rsync.remote_host', usedb=False)
        remote_prefix = self.config.require('rattail.backup', 'rsync.remote_prefix', usedb=False)
        if not remote_prefix.startswith('/'):
            raise RuntimeError("Remote prefix is not absolute path: {}".format(remote_prefix))
        if remote_host:
            remote_prefix = '{}:{}'.format(remote_host, remote_prefix)

        if include:
            prefixes = include
        else:
            prefixes = self.get_path_includes(typ='rsync')
        self.validate_prefixes(prefixes)

        if exclude is None:
            exclude = self.get_path_excludes(typ='rsync')
            if not exclude:
                exclude = self.get_default_excludes()

        for prefix in prefixes:
            if not os.path.exists(prefix):
                log.warning("skipping prefix which doesn't exist locally: %s", prefix)
                continue

            excluded = []
            if exclude:
                excluded = [pattern for pattern in exclude if pattern.startswith(prefix)]
            log.info("running rsync for prefix: %s", prefix)

            parent = os.path.dirname(prefix)
            if parent == '/':
                destination = remote_prefix
                excludes = ['--exclude={}'.format(pattern) for pattern in excluded]

            else: # prefix parent is not root (/) dir
                destination = '{}{}/'.format(remote_prefix, parent)

                # exclusion patterns must apparently be relative in this case,
                # i.e. to the topmost folder being synced
                excludes = []
                for pattern in excluded:
                    pattern = '/'.join(pattern.split('/')[2:])
                    excludes.append('--exclude={}'.format(pattern))

                # and must also create the parent folder on remote side
                cmd = ['rsync', '--dirs', '--perms', '--times', '--group', '--owner', parent, remote_prefix]
                log.debug("rsync command is: %s", cmd)
                if not dry_run:
                    subprocess.check_call(cmd)

            # okay let's rsync
            rsync = ['rsync', '--archive', '--delete-during', '--partial']
            if verbose:
                rsync.append('--progress')
            cmd = rsync + excludes + [prefix, destination]
            log.debug("rsync command is: %s", cmd)
            if not dry_run:
                try:
                    subprocess.check_call(cmd)
                except subprocess.CalledProcessError as error:
                    # nb. rsync exits with code 24 when scenario is
                    # "Partial transfer due to vanished source files"
                    # but that just means all that could be synced, was
                    # synced, though some could not be synced.  we do
                    # not consider that a serious issue, so want to
                    # avoid the outright error here.
                    if error.returncode != 24:
                        raise
                    log.warning("rsync command exited with code 24")

    def make_borg_archives(self, remotes, include=None, exclude=None,
                           tag=None, verbose=False, dry_run=False):

        borg = self.config.get('rattail.backup', 'borg.command', usedb=False)
        if not borg:
            borg = self.config.get('rattail.backup', 'borg', usedb=False,
                                   ignore_ambiguous=True)
            if borg:
                warnings.warn(f"URGENT: instead of 'borg', "
                              f"you should set 'borg.command'",
                              DeprecationWarning, stacklevel=2)
            else:
                borg = os.path.join(os.path.dirname(sys.executable), 'borg')
                if not os.path.exists(borg):
                    borg = subprocess.check_output('which borg')

        if include:
            prefixes = include
        else:
            prefixes = self.get_path_includes(typ='borg')
            if not prefixes:
                prefixes = self.get_path_includes(typ='rsync')
        self.validate_prefixes(prefixes)

        if exclude is None:
            exclude = self.get_path_excludes(typ='borg')
            if not exclude:
                exclude = self.get_path_excludes(typ='rsync')
                if not exclude:
                    exclude = self.get_default_excludes()

        hostname = self.config.get('rattail.backup', 'borg.archive_hostname', usedb=False)
        if not hostname:
            hostname = socket.gethostname()

        for remote in remotes:

            # borg_env
            repo = self.config.require('rattail.backup', 'borg.remote.{}.repo'.format(remote), usedb=False)
            passphrase = self.config.require('rattail.backup', 'borg.remote.{}.passphrase'.format(remote), usedb=False)
            remote_borg = self.config.get('rattail.backup', 'borg.remote.{}.borg'.format(remote), usedb=False)
            borg_env = {
                'HOME': os.environ['HOME'],
                'BORG_REPO': repo,
                'BORG_PASSPHRASE': passphrase,
                'BORG_REMOTE_PATH': remote_borg or 'borg',
            }

            # borg create
            cmd = [borg, 'create',
                   '--compression', 'lz4',
                   '--exclude-caches']
            if verbose:
                cmd.extend([
                    '--verbose',
                    '--list',
                    '--filter', 'AME',
                    '--stats',
                    '--show-rc',
                ])

            for pattern in exclude:
                cmd.extend(['--exclude', pattern])

            if tag:
                # note that in this case our hostname will be immediately
                # followed by a colon instead of hyphen.  this is what "breaks"
                # the typical pattern and ensures this tagged archive will not
                # be pruned under typical command runs.
                archive_name = '::{}:{}-{{now}}'.format(hostname, tag)
            else:
                archive_name = '::{}-{{now}}'.format(hostname)
            cmd.append(archive_name)

            for prefix in prefixes:
                cmd.append(prefix)

            # run `borg create`
            log.debug("borg create command is: %s", cmd)
            if not dry_run:
                try:
                    subprocess.check_call(cmd, env=borg_env)
                except subprocess.CalledProcessError as error:
                    # nb. borg exits with code 1 for "warning" but
                    # this is different than "error" - cf. also
                    # https://borgbackup.readthedocs.io/en/latest/man_intro.html#return-codes
                    if error.returncode != 1:
                        raise
                    log.warning("borg create command exited with code 1")

            # the only thing left is to prune, which if we're tagging, we
            # probably should just skip.  maybe let command line override?
            if tag:
                return

            # generate --keep args for prune, based on config
            keep = []
            keep_defaults = {
                'daily': '7',
                'weekly': '4',
                'monthly': '6',
            }
            for typ, default in keep_defaults.items():
                value = self.config.get('rattail.backup', 'borg.keep.{}'.format(typ))
                if value is None:
                    keep.extend(['--keep-{}'.format(typ), default])
                elif value.isdigit():
                    keep.extend(['--keep-{}'.format(typ), int(value)])
                elif parse_bool(value):
                    keep.extend(['--keep-{}'.format(typ), default])

            # borg prune
            cmd = [borg, 'prune',
                   '--glob-archives', '{}-*'.format(hostname)] + keep
            if verbose:
                cmd.extend([
                    '--list',
                    '--show-rc',
                ])

            log.debug("borg prune command is: %s", cmd)
            if not dry_run:
                subprocess.check_call(cmd, env=borg_env)

    def get_mysql_names(self, include=None, exclude=None, exclude_builtin=True):
        try:
            output = subprocess.check_output([
                'mysql', '--execute', "show databases;",
                '--batch', '--skip-column-names',
            ])
        except Exception as error:
            if isinstance(error, FileNotFoundError):
                # assuming there is no mysql binary
                return []
            raise
        output = output.decode('ascii') # TODO: how to detect this etc.?

        names = output.split('\n')
        names = [name.strip() for name in names if name.strip()]

        if exclude_builtin:
            builtins = [
                'mysql',
                'information_schema',
                'performance_schema',
                'sys',
            ]
            for name in builtins:
                if name in names:
                    names.remove(name)

        return self.filter_names(names, include, exclude)

    def get_mysql_tables(self, database):
        try:
            output = subprocess.check_output([
                'mysql', '--execute', "show tables;",
                '--batch', '--skip-column-names',
                database,
            ])
        except Exception as error:
            if isinstance(error, FileNotFoundError):
                # assuming there is no mysql binary
                return []
            raise
        output = output.decode('ascii') # TODO: how to detect this etc.?

        names = output.split('\n')
        names = [name.strip() for name in names if name.strip()]
        return names

    def get_postgres_names(self, include=None, exclude=None, exclude_builtin=True):
        """
        Returns a list of database names found in PostgreSQL, filtered
        according to the various arguments provided.
        """
        names = []

        # can only use `psql` if it's present, so check first
        with open(os.devnull, 'w') as devnull:
            psql_exists = subprocess.call(['which', 'psql'], stdout=devnull) == 0
        if psql_exists:

            output = subprocess.check_output([
                'sudo', '-u', 'postgres', 'psql', '--list', '--tuples-only',
            ])
            output = output.decode('ascii') # TODO: how to detect this etc.?

            for line in output.split('\n'):
                line = line.strip()
                if line and '|' in line and not line.startswith('|'):
                    name = line.split('|')[0].strip()
                    names.append(name)

        if names and exclude_builtin:
            builtins = [
                'postgres',
                'template0',
                'template1',
            ]
            for name in builtins:
                if name in names:
                    names.remove(name)

        return self.filter_names(names, include, exclude)

    def filter_names(self, names, include, exclude):
        if include:
            names = [name for name in names if name in include]
        elif exclude:
            names = [name for name in names if name not in exclude]
        return names

    def dump_mysql_databases(self, outdir, include=None, exclude=None, tables=False,
                             verbose=False, dry_run=False):
        names = self.get_mysql_names(include=include, exclude=exclude)
        for name in names:
            log.info("dumping mysql db: %s", name)

            # generate mysqldump command
            sql_path = os.path.join(outdir, '{}-mysql.sql'.format(name))
            mysqldump = ['mysqldump']
            if verbose:
                mysqldump.append('--verbose')
            cmd = mysqldump + [name, '--result-file={}'.format(sql_path)]
            log.debug("mysqldump command is: %s", cmd)

            # run mysqldump command
            if not dry_run:
                subprocess.check_call(cmd)
                subprocess.check_call(['gzip', '--force', sql_path])
                gz_path = "{}.gz".format(sql_path)
                log.debug("database dump completed, file is at %s", gz_path)

            if tables:

                # first, forcibly recreate 'tables' dir for this db
                if not dry_run:
                    tablesdir = os.path.join(outdir, 'tables', 'mysql-{}'.format(name))
                    if os.path.exists(tablesdir):
                        shutil.rmtree(tablesdir)
                    os.makedirs(tablesdir)

                # then make a dump of each table reported by 'show tables'
                for table in self.get_mysql_tables(name):
                    log.info("dumping mysql table: %s", table)

                    # generate mysqldump command
                    sql_path = os.path.join(tablesdir, '{}.sql'.format(table))
                    mysqldump = ['mysqldump']
                    if verbose:
                        mysqldump.append('--verbose')
                    cmd = mysqldump + [name, table, '--result-file={}'.format(sql_path)]
                    log.debug("mysqldump command is: %s", cmd)

                    # run mysqldump command
                    if not dry_run:
                        subprocess.check_call(cmd)
                        subprocess.check_call(['gzip', '--force', sql_path])
                        gz_path = "{}.gz".format(sql_path)
                        log.debug("table dump completed, file is at %s", gz_path)

    def dump_postgres_databases(self, outdir, include=None, exclude=None, tables=False,
                                verbose=False, dry_run=False):
        names = self.get_postgres_names(include=include, exclude=exclude)
        for name in names:
            log.info("dumping postgres db: %s", name)

            sql_path = os.path.join(outdir, '{}-postgres.sql'.format(name))
            gz_path = f"{sql_path}.gz"
            tmp_path = f'/tmp/{name}.sql.gz'

            pg_dump = 'sudo -u postgres pg_dump'
            if verbose:
                pg_dump += ' --verbose'
            pg_dump += f" {name} | gzip -c > {tmp_path}"
            cmd = ['bash', '-c', pg_dump]

            logger = log.info if dry_run else log.debug
            logger("pg_dump command is: %s", cmd)

            if not dry_run:
                if os.path.exists(tmp_path):
                    os.remove(tmp_path)
                subprocess.check_call(cmd)
                subprocess.check_call(['chown', 'root:', tmp_path])
                subprocess.check_call(['mv', tmp_path, gz_path])
                log.debug("dump completed, file is at %s", gz_path)

        if tables:
            log.info("TODO: should honor --dump-tables flag for postgres")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck commands
"""

import datetime
import logging

import sqlalchemy as sa
import typer
from typing_extensions import Annotated

from rattail.commands.typer import (make_typer, typer_get_runas_user,
                                    importer_command, typer_eager_imports)
from rattail.commands.importing import ImportCommandHandler
from rattail.time import date_range


log = logging.getLogger(__name__)


trainwreck_typer = make_typer(
    name='trainwreck',
    help="Trainwreck -- Transaction data warehouse")


@trainwreck_typer.command()
@importer_command
def export_trainwreck(
        ctx: typer.Context,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for database engine to be used as the \"target\" "
                         "Trainwreck DB, i.e. where data will be exported.  This key must "
                         "be defined in the [trainwreck.db] section of your config file.")] = None,
        **kwargs
):
    """
    Export data to another Trainwreck database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_trainwreck.from_trainwreck.export')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'dbkey': dbkey}
    handler.run(kwargs, progress=progress)


@trainwreck_typer.command()
@importer_command
def import_self(
        ctx: typer.Context,
        **kwargs
):
    """
    Self-update for Trainwreck data
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_trainwreck.from_self.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    handler.run(kwargs, progress=progress)


@trainwreck_typer.command()
@importer_command
def import_trainwreck(
        ctx: typer.Context,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for database engine to be used as the "
                         "Trainwreck \"host\", i.e. the source of the data to be "
                         "imported.  This key must be defined in the [trainwreck.db] "
                         "section of your config file.")] = None,
        **kwargs
):
    """
    Import data from another Trainwreck database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_trainwreck.from_trainwreck.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'dbkey': dbkey}
    handler.run(kwargs, progress=progress)


@trainwreck_typer.command()
def prune(
        ctx: typer.Context,
        after: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Date *after* which all data should be pruned.  If set, no data "
                         "will be pruned from this or earlier dates.  If not set, there will be "
                         "no lower boundary for the prune.")] = None,
        before: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Date *before* which all data should be pruned.  If set, no data "
                         "will be pruned from this or later dates.  If not set, will assume a "
                         "default of \"today\".")] = None,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for the Trainwreck database engine to be used, i.e. "
                         "from which data should be pruned.  This key must be defined in the "
                         "[trainwreck.db] section of your config file.")] = 'default',
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Prune some dates from a Trainwreck database
    """
    from rattail.db.util import finalize_session

    config = ctx.parent.rattail_config
    app = config.get_app()
    trainwreck = app.get_trainwreck_handler()
    progress = ctx.parent.rattail_progress

    session = trainwreck.make_session(dbkey=dbkey)
    model = trainwreck.get_model()

    if after:
        start_date = after.date() + datetime.timedelta(days=1)
    else:
        # default is earliest available date
        end_time = session.query(sa.func.min(model.Transaction.end_time)).scalar()
        start_date = app.localtime(end_time, from_utc=True).date()

    if before:
        end_date = before.date() - datetime.timedelta(days=1)
    else:
        # default is yesterday
        end_date = app.yesterday()

    dates = list(date_range(start_date, end_date))

    for date in dates:
        log.debug("pruning from '%s' Trainwreck for %s", dbkey, date)

        start_time = datetime.datetime.combine(date, datetime.time(0))
        start_time = app.localtime(start_time)

        end_time = datetime.datetime.combine(date + datetime.timedelta(days=1), datetime.time(0))
        end_time = app.localtime(end_time)

        trainwreck.prune_date(session, date, start_time, end_time, progress=progress)

    finalize_session(session, dry_run=dry_run)


# discover more commands
typer_eager_imports(trainwreck_typer)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Autocomplete Handler for People
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm

from rattail.autocomplete import Autocompleter
from rattail.autocomplete.base import PhoneMagicMixin
from rattail.db import model


class PersonAutocompleter(Autocompleter):
    """
    Autocompleter for People
    """
    autocompleter_key = 'people'
    model_class = model.Person
    autocomplete_fieldname = 'display_name'


class PersonEmployeeAutocompleter(Autocompleter):
    """
    Autocompleter for People, but restricted to return only results
    for people who are (or have been) an employee.
    """
    autocompleter_key = 'people.employees'
    model_class = model.Person
    autocomplete_fieldname = 'display_name'

    def restrict_autocomplete_query(self, session, query, **kwargs):
        model = self.model
        query = query.join(model.Employee)
        return query


class PersonNewOrderAutocompleter(PhoneMagicMixin, Autocompleter):
    """
    Special "new order" autocompleter for people.

    We set it apart with a different key (``'people.neworder'``) so
    that you can override it independently of other person
    autocompleters.

    But the default logic for this one is a bit special too, in that
    it will try to search for *either* phone number *or* person name.
    If the search term includes at least 4 digits then it is
    considered to be a phone number search; otherwise it will be
    considered a name search.
    """
    autocompleter_key = 'people.neworder'
    model_class = model.Person
    autocomplete_fieldname = 'display_name'
    phone_model_class = model.PersonPhoneNumber

    def make_base_query(self, session):
        query = super(PersonNewOrderAutocompleter, self).make_base_query(session)
        return query.options(orm.joinedload(self.model_class.emails))

    def get_autocomplete_results(self, data):
        results = []
        for contact in data:
            name = contact.display_name

            values = [name]

            phone = contact.first_phone()
            if phone:
                values.append(phone.number)

            email = contact.first_email()
            if email:
                values.append(email.address)

            results.append({'value': contact.uuid,
                            'label': ' / '.join(values),
                            'display': name})
        return results



# -*- coding: utf-8; -*-
"""add batch description

Revision ID: 539c2df56562
Revises: 1f21b32fd2a7
Create Date: 2017-08-09 19:38:28.956119

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '539c2df56562'
down_revision = u'1f21b32fd2a7'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch*
    op.add_column('batch_handheld', sa.Column('description', sa.String(length=255), nullable=True))
    op.add_column('batch_inventory', sa.Column('description', sa.String(length=255), nullable=True))
    op.add_column('batch_pricing', sa.Column('description', sa.String(length=255), nullable=True))
    op.add_column('label_batch', sa.Column('description', sa.String(length=255), nullable=True))
    op.add_column('purchase_batch', sa.Column('description', sa.String(length=255), nullable=True))
    op.add_column('vendor_catalog', sa.Column('description', sa.String(length=255), nullable=True))
    op.add_column('vendor_invoice', sa.Column('description', sa.String(length=255), nullable=True))


def downgrade():

    # batch*
    op.drop_column('vendor_invoice', 'description')
    op.drop_column('vendor_catalog', 'description')
    op.drop_column('purchase_batch', 'description')
    op.drop_column('label_batch', 'description')
    op.drop_column('batch_pricing', 'description')
    op.drop_column('batch_inventory', 'description')
    op.drop_column('batch_handheld', 'description')



# -*- coding: utf-8; -*-
"""add pending_customer table

Revision ID: 5a256a77e6d0
Revises: 51f7773bb07a
Create Date: 2021-10-06 16:35:39.749555

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '5a256a77e6d0'
down_revision = u'51f7773bb07a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # pending_customer
    op.create_table('pending_customer',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('id', sa.String(length=20), nullable=True),
                    sa.Column('first_name', sa.String(length=50), nullable=True),
                    sa.Column('middle_name', sa.String(length=50), nullable=True),
                    sa.Column('last_name', sa.String(length=50), nullable=True),
                    sa.Column('display_name', sa.String(length=100), nullable=True),
                    sa.Column('phone_number', sa.String(length=20), nullable=True),
                    sa.Column('phone_type', sa.String(length=15), nullable=True),
                    sa.Column('email_address', sa.String(length=255), nullable=True),
                    sa.Column('email_type', sa.String(length=15), nullable=True),
                    sa.Column('address_street', sa.String(length=100), nullable=True),
                    sa.Column('address_street2', sa.String(length=100), nullable=True),
                    sa.Column('address_city', sa.String(length=60), nullable=True),
                    sa.Column('address_state', sa.String(length=2), nullable=True),
                    sa.Column('address_zipcode', sa.String(length=10), nullable=True),
                    sa.Column('address_type', sa.String(length=15), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['user_uuid'], [u'user.uuid'], name=u'pending_customer_fk_user'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # batch_custorder
    op.add_column(u'batch_custorder', sa.Column('pending_customer_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'batch_custorder_fk_pending_customer', 'batch_custorder', 'pending_customer', ['pending_customer_uuid'], ['uuid'])

    # custorder
    op.add_column(u'custorder', sa.Column('pending_customer_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'custorder_fk_pending_customer', 'custorder', 'pending_customer', ['pending_customer_uuid'], ['uuid'])


def downgrade():

    # custorder
    op.drop_constraint(u'custorder_fk_pending_customer', 'custorder', type_='foreignkey')
    op.drop_column(u'custorder', 'pending_customer_uuid')

    # batch_custorder
    op.drop_constraint(u'batch_custorder_fk_pending_customer', 'batch_custorder', type_='foreignkey')
    op.drop_column(u'batch_custorder', 'pending_customer_uuid')

    # pending_customer
    op.drop_table('pending_customer')



# -*- coding: utf-8; -*-
"""add invoice_date for receiving row

Revision ID: 093db8c16e7d
Revises: 0d1c2362b8b7
Create Date: 2023-08-07 13:40:04.545706

"""

# revision identifiers, used by Alembic.
revision = '093db8c16e7d'
down_revision = '0d1c2362b8b7'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # purchase_item
    op.add_column('purchase_item', sa.Column('invoice_date', sa.Date(), nullable=True))

    # batch_purchase_row
    op.add_column('batch_purchase_row', sa.Column('invoice_date', sa.Date(), nullable=True))


def downgrade():

    # batch_purchase_row
    op.drop_column('batch_purchase_row', 'invoice_date')

    # purchase_item
    op.drop_column('purchase_item', 'invoice_date')



# -*- coding: utf-8; -*-
"""add custorder_item.flagged

Revision ID: fa3aec1556bc
Revises: 86b00e66b83e
Create Date: 2023-09-10 14:33:56.552430

"""

# revision identifiers, used by Alembic.
revision = 'fa3aec1556bc'
down_revision = '86b00e66b83e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # custorder_item
    op.add_column('custorder_item', sa.Column('special_order', sa.Boolean(), nullable=True))
    op.add_column('custorder_item', sa.Column('flagged', sa.Boolean(), nullable=True))
    op.add_column('custorder_item', sa.Column('contact_attempts', sa.Integer(), nullable=True))
    op.add_column('custorder_item', sa.Column('last_contacted', sa.DateTime(), nullable=True))

    # batch_custorder_row
    op.add_column('batch_custorder_row', sa.Column('special_order', sa.Boolean(), nullable=True))


def downgrade():

    # batch_custorder_row
    op.drop_column('batch_custorder_row', 'special_order')

    # custorder_item
    op.drop_column('custorder_item', 'last_contacted')
    op.drop_column('custorder_item', 'contact_attempts')
    op.drop_column('custorder_item', 'flagged')
    op.drop_column('custorder_item', 'special_order')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail mailmon`` command
"""

import sys
from enum import Enum
from pathlib import Path

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


class ServiceAction(str, Enum):
    start = 'start'
    stop = 'stop'


@rattail_typer.command()
def mailmon(
        ctx: typer.Context,
        action: Annotated[
            ServiceAction,
            typer.Argument(help="Action to perform for the service.")] = ...,
        pidfile: Annotated[
            Path,
            typer.Option('--pidfile', '-p',
                         help="Path to PID file.")] = '/var/run/rattail/mailmon.pid',
):
    """
    Manage the mail monitor daemon
    """
    from rattail.mailmon.daemon import MailMonitorDaemon

    config = ctx.parent.rattail_config
    daemon = MailMonitorDaemon(pidfile, config=config)
    if action == 'stop':
        daemon.stop()
    else: # start
        try:
            daemon.start(daemonize=False)
        except KeyboardInterrupt:
            sys.stderr.write("Interrupted.\n")


class MailMonitorCommand(Subcommand):
    """
    Interacts with the mail monitor service; called as ``rattail
    mailmon``.  This command expects a subcommand; one of the
    following:

    * ``rattail mailmon start``
    * ``rattail mailmon stop``
    """
    name = 'mailmon'
    description = "Manage the mail monitor daemon"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')

        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        parser.add_argument('-p', '--pidfile', metavar='PATH', default='/var/run/rattail/mailmon.pid',
                            help="Path to PID file.")

    def run(self, args):
        from rattail.mailmon.daemon import MailMonitorDaemon

        daemon = MailMonitorDaemon(args.pidfile, config=self.config)
        if args.subcommand == 'stop':
            daemon.stop()
        else: # start
            try:
                daemon.start(daemonize=False)
            except KeyboardInterrupt:
                self.stderr.write("Interrupted.\n")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Typer-based command utilities
"""

import datetime
import inspect
import logging
from pathlib import Path
from typing import List, Optional

import makefun
import typer
from typing_extensions import Annotated
from wuttjamaican.util import load_entry_points

from rattail.config import make_config
from rattail.db.config import configure_versioning
from rattail.progress import ConsoleProgress


def make_typer(**kwargs):
    """
    Create a Typer command instance, per Rattail conventions.

    :param epname: Setuptools entry point name for the command.
    """
    kwargs.setdefault('callback', typer_callback)
    return typer.Typer(**kwargs)


def typer_eager_imports(
        group: [typer.Typer, str]):
    """
    Eagerly import all modules which are registered as having commands
    belonging to the given group.

    :param group: Typer group command, or the name of one.
    """
    if isinstance(group, typer.Typer):
        group = group.info.name
    load_entry_points(f'{group}.typer_imports')


def typer_callback(
        ctx: typer.Context,

        ##############################

        # nb. these first args are defined in wuttjamaican

        config_paths: Annotated[
            Optional[List[Path]],
            typer.Option('--config', '-c',
                         exists=True,
                         help="Config path (may be specified more than once)")] = None,

        plus_config_paths: Annotated[
            Optional[List[Path]],
            typer.Option('--plus-config',
                         exists=True,
                         help="Extra configs to load in addition to normal config")] = None,

        progress: Annotated[
            bool,
            typer.Option('--progress', '-P',
                          help="Report progress when relevant")] = False,

        # # fn = click.option('-V', '--version',
        # #                   is_flag=True,
        # #                   help="Show program version"
        # #                   )(fn)

        # fn = click.option('--stdout',
        #                   type=click.File('wb'),
        #                   help="Optional file to which STDOUT should be written"
        #                   )(fn)

        # fn = click.option('--stderr',
        #                   type=click.File('wb'),
        #                   help="Optional file to which STDERR should be written"
        #                   )(fn)

        ##############################

        # nb. the rest of these args are defined in rattail

        no_init: Annotated[
            bool,
            typer.Option('--no-init')] = False,

        no_extend_config: Annotated[
            bool,
            typer.Option('--no-extend-config')] = False,

        verbose: Annotated[
            bool,
            typer.Option('--verbose')] = False,

        progress_socket: Annotated[
            str,
            typer.Option(help="Optional socket (e.g. localhost:8487) to which progress "
                         "info should be written.")] = None,

        runas_username: Annotated[
            str,
            typer.Option('--runas', '-R',
                         help="Optional username to impersonate when running the command.  "
                         "This is only relevant for / used by certain commands.")] = None,

        versioning: Annotated[
            bool,
            typer.Option('--versioning',
                         help="Force *enable* of data versioning.  If set, then --no-versioning "
                         "cannot also be set.  If neither is set, config will determine whether "
                         "or not data versioning should be enabled.")] = False,

        no_versioning: Annotated[
            bool,
            typer.Option('--no-versioning',
                         help="Force *disable* of data versioning.  If set, then --versioning "
                         "cannot also be set.  If neither is set, config will determine whether "
                         "or not data versioning should be enabled.")] = False,

):
    """
    Generic callback for use with top-level commands.
    """
    config = make_cli_config(ctx)

    ctx.rattail_config = config
    ctx.rattail_progress = ConsoleProgress if progress else None


def make_cli_config(ctx):
    """
    Titeship (custom Rattail system)
    """
    kwargs = ctx.params

    # if args say not to "init" then we make a sort of empty config
    if kwargs['no_init']:
        config = make_config([], extend=False, versioning=False)

    else: # otherwise we make a proper config, and maybe turn on versioning
        logging.basicConfig()
        config = make_config(files=kwargs['config_paths'] or None,
                             plus_files=kwargs['plus_config_paths'] or None,
                             extend=not kwargs['no_extend_config'],
                             versioning=False)
        if kwargs['versioning']:
            configure_versioning(config, force=True)
        elif not kwargs['no_versioning']:
            configure_versioning(config)

    # import our primary data model now, just in case it hasn't fully been
    # imported yet.  this it to be sure association proxies and the like
    # are fully wired up in the case of extensions
    # TODO: what about web apps etc.? i guess i was only having the problem
    # for some importers, e.g. basic CORE API -> Rattail w/ the schema
    # extensions in place from rattail-corepos
    try:
        config.get_model()
    except ImportError:
        pass

    return config


def typer_get_runas_user(ctx, session=None):
    """
    Convenience function to get the "runas" User object for the
    current command.

    Uses :meth:`rattail.app.AppHandler.get_runas_user()` under the
    hood, but the ``--runas`` command line param provides the default
    username.
    """
    config = ctx.parent.rattail_config
    app = config.get_app()
    return app.get_runas_user(session=session,
                              username=ctx.parent.params['runas_username'])


def importer_command_template(

        # model names (aka importer keys)
        models: Annotated[
            Optional[List[str]],
            typer.Argument(help="Which data models to import.  If you specify any, then only "
                           "data for those models will be imported.  If you do not specify "
                           "any, then all *default* models will be imported.")] = None,

        # list models
        list_all_models: Annotated[
            bool,
            typer.Option('--list-all-models', '-l',
                         help="List all available models and exit.")] = False,
        list_default_models: Annotated[
            bool,
            typer.Option('--list-default-models',
                         help="List the default models and exit.")] = False,

        # make batches
        make_batches: Annotated[
            bool,
            typer.Option('--make-batches',
                         help="If specified, make new Import / Export Batches instead of "
                         "performing an actual (possibly dry-run) import.")] = False,

        # key / fields / exclude
        key: Annotated[
            str,
            typer.Option('--key',
                         help="List of fields (comma-separated string) which should be "
                         "used as \"primary key\" for the import.")] = None,
        fields: Annotated[
            str,
            typer.Option('--fields',
                         help="List of fields (comma-separated string) which should be "
                         "included in the import.  If this parameter is specified, then "
                         "any field not listed here, would be *excluded* regardless of "
                         "the --exclude-fields parameter.")] = None,
        exclude_fields: Annotated[
            str,
            typer.Option('--exclude-fields',
                         help="List of fields (comma-separated string) which should be "
                         "excluded from the import.  Any field not listed here, would be "
                         "included (or not) depending on the --fields parameter and/or "
                         "default importer behavior.")] = None,

        # fuzzy fields
        fuzzy_fields: Annotated[
            str,
            typer.Option(help="List of fields (comma-separated string) for which diff "
                         "comparison should be \"fuzzy\".  This is intended for "
                         "timestamps and similar values which vary in granularity "
                         "between systems.")] = None,
        fuzz_factor: Annotated[
            int,
            typer.Option(help="Numeric value for use with --fuzzy-fields.  For "
                         "timestamp fields, this refers to the number of seconds "
                         "by which values are allowed to differ and still be "
                         "considered a match.")] = 1,

        # allow create?
        create: Annotated[
            bool,
            typer.Option(help="Allow new records to be created during the import.")] = True,
        max_create: Annotated[
            int,
            typer.Option(help="Maximum number of records which may be created, after which a "
                         "given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # allow update?
        update: Annotated[
            bool,
            typer.Option(help="Allow existing records to be updated during the import.")] = True,
        max_update: Annotated[
            int,
            typer.Option(help="Maximum number of records which may be updated, after which a "
                         "given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # allow delete?
        delete: Annotated[
            bool,
            typer.Option(help="Allow records to be deleted during the import.")] = False,
        max_delete: Annotated[
            int,
            typer.Option(help="Maximum number of records which may be deleted, after which a "
                         "given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # max total changes, per model
        max_total: Annotated[
            int,
            typer.Option(help="Maximum number of *any* record changes which may occur, after which "
                         "a given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # date ranges
        start_date: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Optional (inclusive) starting point for date range, by which host "
                         "data should be filtered.  Only used by certain importers.")] = None,
        end_date: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Optional (inclusive) ending point for date range, by which host "
                         "data should be filtered.  Only used by certain importers.")] = None,
        year: Annotated[
            int,
            typer.Option(help="Optional year, by which data should be filtered.  Only used "
                         "by certain importers.")] = None,

        # TODO: deprecate --batch, replace with --batch-size ?
        # batch size
        batch_size: Annotated[
            int,
            typer.Option('--batch',
                         help="Split work to be done into batches, with the specified number of "
                         "records in each batch.  Or, set this to 0 (zero) to disable batching. "
                         "Implementation for this may vary somewhat between importers.")] = 200,

        # collect changes for processing
        collect_changes: Annotated[
            bool,
            typer.Option(help="Collect changes for processing at the end of the run.  "
                         "This is required for warning diff emails, but disabling it "
                         "can help to cut down on memory usage during the run.")] = True,

        # email diff warnings
        warnings: Annotated[
            bool,
            typer.Option('--warnings', '-W',
                         help="Set this flag if you expect a \"clean\" import, and wish for any "
                         "changes which do occur to be processed further and/or specially.  The "
                         "behavior of this flag is ultimately up to the import handler, but the "
                         "default is to send an email notification.")] = False,
        max_diffs: Annotated[
            int,
            typer.Option(help="Maximum number of \"diffs\" to display per warning type, in a "
                         "warning email.  Only used if --warnings is in effect.")] = None,

        # dry run?
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.  "
                         "Note that this flag is ignored if --make-batches is specified.")] = False,
):
    """
    Stub function which provides a common param signature; used with
    :func:`importer_command`.
    """


def importer_command(fn):
    """
    Decorator for import/export commands.  Adds common params based on
    :func:`importer_command_template`.
    """
    original_sig = inspect.signature(fn)
    reference_sig = inspect.signature(importer_command_template)

    params = list(original_sig.parameters.values())
    for i, param in enumerate(reference_sig.parameters.values()):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)


def file_exporter_command_template(
        output_dir: Annotated[
            Path,
            typer.Option(help="Directory to which output files should be written.  "
                         "Note that this is a *required* parameter.")] = ...,
):
    """
    Stub function to provide signature for exporter commands which
    produce data file(s) as output.  Used with
    :func:`file_exporter_command`.
    """


def file_exporter_command(fn):
    """
    Decorator for file export commands.  Adds common params based on
    :func:`file_exporter_command_template`.
    """
    original_sig = inspect.signature(fn)
    plain_import_sig = inspect.signature(importer_command_template)
    file_export_sig = inspect.signature(file_exporter_command_template)
    desired_params = (
        list(plain_import_sig.parameters.values())
        + list(file_export_sig.parameters.values()))

    params = list(original_sig.parameters.values())
    for i, param in enumerate(desired_params):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)


def file_importer_command_template(
        input_dir: Annotated[
            Path,
            typer.Option(exists=True,
                         help="Directory from which input files should be read.  "
                         "Note that this is a *required* parameter.")] = ...,
):
    """
    Stub function to provide signature for importer commands which
    require data file(s) as input.  Used with
    :func:`file_importer_command`.
    """


def file_importer_command(fn):
    """
    Decorator for file import commands.  Adds common params based on
    :func:`file_importer_command_template`.
    """
    original_sig = inspect.signature(fn)
    plain_import_sig = inspect.signature(importer_command_template)
    file_import_sig = inspect.signature(file_importer_command_template)
    desired_params = (
        list(plain_import_sig.parameters.values())
        + list(file_import_sig.parameters.values()))

    params = list(original_sig.parameters.values())
    for i, param in enumerate(desired_params):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
New Report Feature
"""

from __future__ import unicode_literals, absolute_import

import re

from rattail.features import Feature

import colander


class NewReportFeature(Feature):
    """
    New Report Feature
    """
    feature_key = 'new-report'
    feature_title = "New Report"

    def make_schema(self, **kwargs):

        class Schema(colander.MappingSchema):

            name = colander.SchemaNode(colander.String())

            description = colander.SchemaNode(colander.String())

        return Schema(**kwargs)

    def get_defaults(self):
        return {
            'name': "Latest Widgets",
            'description': "Shows all the latest widgets.",
        }

    def generate(self, **kwargs):
        context = dict(kwargs)
        context['app_package'] = self.config.app_package()
        context['app_slug'] = context['app_package'] # TODO
        context['code_name'] = re.sub(r'[\s-]', '_', context['name']).lower()
        context['cap_name'] = context['name'].replace(' ', '')

        return self.generate_mako('/new-report/instructions.mako', **context)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Users & Permissions
"""

import datetime
import warnings

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.associationproxy import association_proxy

from rattail.db.model import Base, uuid_column, getset_factory, Person
 

class Role(Base):
    """
    Represents a role within the system; used to manage permissions.
    """
    __tablename__ = 'role'
    __table_args__ = (
        sa.UniqueConstraint('name', name='role_uq_name'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    name = sa.Column(sa.String(length=100), nullable=False, doc="""
    Name for the role.  Each role must have a name, which must be unique.
    """)

    adminish = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that the role is "admin-ish" - i.e. only users who
    belong to the true Administrator role, should be allowed to
    (un)assign users to this role.
    """)

    session_timeout = sa.Column(sa.Integer(), nullable=True, doc="""
    Optional session timeout value for the role, in seconds.  If this is set to
    zero, the role's users will have no session timeout.  A value of ``None``
    means the role has no say in the timeout.
    """)

    notes = sa.Column(sa.Text(), nullable=True, doc="""
    Any arbitrary notes for the role.
    """)

    sync_me = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that the Role - its primary attributes, and list
    of permissions - should be synced across all nodes.

    So if set, when the role changes at one node then that change
    should propagate to all other nodes.

    Note that this does *not* include the user list by default; see
    :attr:`sync_users` to add that.

    Note that if this flag is set, the role will be synced to *all*
    nodes regardless of node type.  See also :attr:`node_type`.
    """)

    sync_users = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that the user list for the role should be synced
    across all nodes.  This has no effect unless :attr:`sync_me` is
    also set.

    Note that if this flag is set, the role's user list will be synced
    to *all* nodes regardless of node type.  See also
    :attr:`node_type`.
    """)

    node_type = sa.Column(sa.String(length=100), nullable=True, doc="""
    Type of node for which this role is applicable.  This is probably
    only useful if the :attr:`sync_me` flag is set.

    If set, this value must match a node's configured type, or else it
    will be ignored by that node.  See also
    :meth:`~rattail.config.RattailConfig.node_type()` for how a node's
    type is determined.  If there is no value set for this field then
    the role will be honored by all nodes in which it exists (which is
    just one unless ``sync_me`` is set, in which case all nodes would
    have it).

    It is useful in combination with ``sync_me`` in that it allows a
    certain role to be "global" (synced) and yet only be "effective"
    for certain nodes.  Probably the most common scenario is where you
    have a host node and several store nodes, and you want to manage
    the store roles "globally" but avoid granting unwanted access to
    the host node.  So you'd set the ``sync_me`` flag but also set
    ``node_type`` to e.g. ``'store'``.
    """)

    _users = orm.relationship(
        'UserRole',
        cascade='all, delete-orphan',
        back_populates='role',
        cascade_backrefs=False)

    users = association_proxy(
        '_users', 'user',
        creator=lambda u: UserRole(user=u),
        getset_factory=getset_factory)

    def __str__(self):
        return self.name or ''


class Permission(Base):
    """
    Represents permission a role has to do a particular thing.
    """
    __tablename__ = 'permission'
    __table_args__ = (
        sa.ForeignKeyConstraint(['role_uuid'], ['role.uuid'], name='permission_fk_role'),
        )
    # __versioned__ = {}

    role_uuid = sa.Column(sa.String(length=32), primary_key=True)
    permission = sa.Column(sa.String(length=254), primary_key=True)

    def __str__(self):
        return self.permission or ''


Role._permissions = orm.relationship(
    Permission, backref='role',
    cascade='save-update, merge, delete, delete-orphan')

Role.permissions = association_proxy(
    '_permissions', 'permission',
    creator=lambda p: Permission(permission=p),
    getset_factory=getset_factory)


class User(Base):
    """
    Represents a user of the system.

    This may or may not correspond to a real person, i.e. some users may exist
    solely for automated tasks.
    """
    __tablename__ = 'user'
    __table_args__ = (
        sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='user_fk_person'),
        sa.UniqueConstraint('username', name='user_uq_username'),
        sa.Index('user_ix_person', 'person_uuid'),
    )
    __versioned__ = {'exclude': ['password', 'salt', 'last_login']}

    uuid = uuid_column()
    username = sa.Column(sa.String(length=25), nullable=False)
    password = sa.Column(sa.String(length=60))
    salt = sa.Column(sa.String(length=29))

    person_uuid = sa.Column(sa.String(length=32))
    person = orm.relationship(
        Person,
        uselist=False,
        doc="""
        Reference to the person whose user account this is.
        """,
        backref=orm.backref(
            'users',
            cascade_backrefs=False,
            doc="""
            List of user accounts for the person.  Typically there is
            only one user account per person, but technically multiple
            are allowed.
            """))

    active = sa.Column(sa.Boolean(), nullable=False, default=True, doc="""
    Whether the user is active, e.g. allowed to log in via the UI.
    """)

    active_sticky = sa.Column(sa.Boolean(), nullable=True, doc="""
    Optional flag, motivation behind which is as follows: If you import user
    accounts from another system, esp. on a regular basis, you might be keeping
    the :attr:`active` flag in sync along with that.  But in some cases you
    might want to *not* keep the active flag in sync, for certain accounts.
    Hence this "active sticky" flag, which may be used to mark certain accounts
    as off-limits from the general active flag sync.
    """)

    prevent_password_change = sa.Column(sa.Boolean(), nullable=True, doc="""
    If set, this user cannot change their own password, *and* the
    password is not editable when e.g. a manager edits this user
    record.  So if set, only root can change this user's password.
    """)

    local_only = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the user account is somehow specific to the "local" app
    node etc. and should not be synced elsewhere.
    """)

    last_login = sa.Column(sa.DateTime(), nullable=True, doc="""
    Timestamp when user last logged into the system.
    """)

    def __str__(self):
        if self.person and str(self.person):
            return str(self.person)
        return self.username or ''

    @property
    def display_name(self):
        """
        Display name for the user.
        
        Returns :attr:`Person.display_name` if available; otherwise returns
        :attr:`username`.
        """
        if self.person and self.person.display_name:
            return self.person.display_name
        return self.username

    @property
    def employee(self):
        """
        DEPRECATED

        Reference to the :class:`Employee` associated with the user, if any.
        """
        warnings.warn("user.employee is deprecated; please use "
                      "app.get_employee(user) instead",
                      DeprecationWarning, stacklevel=2)
        if self.person:
            return self.person.employee

    def get_short_name(self):
        """
        Returns "short name" for the user.  This is for convenience of mobile
        view, at least...
        """
        warnings.warn("user.get_short_name() is deprecated; please use "
                      "AuthHandler.get_short_display_name(user) instead",
                      DeprecationWarning, stacklevel=2)

        # TODO: this should reference employee.short_name
        employee = self.employee
        if employee and employee.display_name:
            return employee.display_name

        person = self.person
        if person:
            if person.first_name and person.last_name:
                return "{} {}.".format(person.first_name, person.last_name[0])
            if person.first_name:
                return person.first_name

        return self.username

    def get_email_address(self):
        """
        DEPRECATED

        Returns the primary email address for the user (as unicode string), or
        ``None``.  Note that currently there is no direct association between a
        User and an EmailAddress, so the Person and Customer relationships are
        navigated in an attempt to locate an address.
        """
        warnings.warn("user.get_email_address() is deprecated; please "
                      "use app.get_contact_email_address(user) instead",
                      DeprecationWarning, stacklevel=2)
        if self.person:
            if self.person.email:
                return self.person.email.address
            for customer in self.person.customers:
                if customer.email:
                    return customer.email.address

    @property
    def email_address(self):
        """
        DEPRECATED

        Convenience attribute which invokes :meth:`get_email_address()`.

        .. note::
           The implementation of this may change some day, e.g. if the User is
           given an association to EmailAddress in the data model.
        """
        warnings.warn("user.email_address is deprecated; please "
                      "use app.get_contact_email_address(user) instead",
                      DeprecationWarning, stacklevel=2)
        return self.get_email_address()

    def is_admin(self):
        """
        Convenience method to determine if current user is a member of the
        Administrator role.
        """
        from rattail.db.auth import administrator_role

        session = orm.object_session(self)
        return administrator_role(session) in self.roles

    def record_event(self, type_code, **kwargs):
        kwargs['type_code'] = type_code
        self.events.append(UserEvent(**kwargs))


Person.make_proxy(User, 'person', 'first_name')
Person.make_proxy(User, 'person', 'last_name')
Person.make_proxy(User, 'person', 'display_name')


class UserRole(Base):
    """
    Represents the association between a :class:`User` and a :class:`Role`.
    """
    __tablename__ = 'user_x_role'
    __table_args__ = (
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='user_x_role_fk_user'),
        sa.ForeignKeyConstraint(['role_uuid'], ['role.uuid'], name='user_x_role_fk_role'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    user_uuid = sa.Column(sa.String(length=32), nullable=False)

    role_uuid = sa.Column(sa.String(length=32), nullable=False)
    role = orm.relationship(Role, back_populates='_users')


User._roles = orm.relationship(
    UserRole, backref='user',
    cascade='all, delete-orphan')

User.roles = association_proxy(
    '_roles', 'role',
    creator=lambda r: UserRole(role=r),
    getset_factory=getset_factory)


class UserEvent(Base):
    """
    Represents an event associated with a user.
    """
    __tablename__ = 'user_event'
    __table_args__ = (
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'],
                                name='user_event_fk_user'),
    )

    uuid = uuid_column()

    user_uuid = sa.Column(sa.String(length=32), nullable=False)
    user = orm.relationship(
        User,
        doc="""
        Reference to the user whose event this is.
        """,
        backref=orm.backref(
            'events',
            cascade='all, delete-orphan',
            cascade_backrefs=False,
            doc="""
            Sequence of events for the user.
            """))

    type_code = sa.Column(sa.Integer(), nullable=False, doc="""
    Type code for the event.
    """)

    occurred = sa.Column(sa.DateTime(), nullable=True, default=datetime.datetime.utcnow, doc="""
    Timestamp at which the event occurred.
    """)


class UserAPIToken(Base):
    """
    User authentication token for use with Tailbone API
    """
    __tablename__ = 'user_api_token'
    __table_args__ = (
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'],
                                name='user_api_token_fk_user'),
    )
    __versioned__ = {}
    model_title = "API Token"
    model_title_plural = "API Tokens"

    uuid = uuid_column()

    user_uuid = sa.Column(sa.String(length=32), nullable=False, doc="""
    Reference to the User associated with the token.
    """)
    user = orm.relationship(
        'User',
        doc="""
        Reference to the User associated with the token.
        """,
        backref=orm.backref(
            'api_tokens',
            cascade_backrefs=False,
            order_by='UserAPIToken.created',
            doc="""
            List of API tokens for the user.
            """))

    description = sa.Column(sa.String(length=255), nullable=False, doc="""
    Description of the token.
    """)

    token_string = sa.Column(sa.String(length=255), nullable=False, doc="""
    Token string, to be used by API clients.
    """)

    created = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    Date/time when the token was created.
    """)

    def __str__(self):
        return self.description or ""



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail make-user`` command
"""

import sys
from enum import Enum
from getpass import getpass

import typer
from typing_extensions import Annotated
from wuttjamaican.util import parse_list

from .base import rattail_typer, Subcommand


class UserSystem(str, Enum):
    rattail = 'rattail'
    windows = 'windows'


@rattail_typer.command()
def make_user(
        ctx: typer.Context,
        username: Annotated[
            str,
            typer.Argument(help="Username for the new user.")] = ...,
        system: Annotated[
            UserSystem,
            typer.Option(help="System in which to create the new user.")] = 'rattail',
        admin: Annotated[
            bool,
            typer.Option('--admin', '-A',
                         help="Whether the new user should have admin rights within "
                         "the system (if applicable).")] = False,
        password: Annotated[
            str,
            typer.Option(help="Password to set for the new user.  If not specified, "
                         "you may be prompted for one.")] = None,
        no_password: Annotated[
            bool,
            typer.Option('--no-password',
                         help="Do not ask for, or try to set, a password for the new user.")] = False,
        full_name: Annotated[
            str,
            typer.Option(help="Full (display) name for the new user (if applicable).")] = None,
        comment: Annotated[
            str,
            typer.Option(help="Comment string for the new user (if applicable).")] = None,
        groups: Annotated[
            str,
            typer.Option(help="Optional list (comma-separated string) of groups/roles "
                         "to which the new user should be assigned.")] = None,
):
    """
    Create a new user account in a given system
    """
    config = ctx.parent.rattail_config

    mkuser = None
    if system == 'rattail':
        mkuser = mkuser_rattail
    elif system == 'windows':
        mkuser = mkuser_windows

    if mkuser:
        if mkuser(config, ctx.params):
            sys.stdout.write(f"created new user in '{system}' system: {username}\n")
    else:
        sys.stderr.write(f"don't know how to make user for '{system}' system\n")
        sys.exit(1)


class MakeUser(Subcommand):
    """
    Create a new user account in a given system
    """
    name = 'make-user'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('username',
                            help="Username for the new user.")
        parser.add_argument('--system', default='rattail',
                            help="System in which to create the new user; defaults to "
                            "rattail; must be one of: rattail, windows")
        parser.add_argument('-A', '--admin', action='store_true',
                            help="Whether the new user should have admin rights within "
                            "the system (if applicable).")
        parser.add_argument('--password',
                            help="Password to set for the new user.  If not specified, "
                            "you may be prompted for one.")
        parser.add_argument('--no-password', action='store_true',
                            help="Do not ask for, or try to set, a password for the new user.")
        parser.add_argument('--full-name',
                            help="Full (display) name for the new user (if applicable).")
        parser.add_argument('--comment',
                            help="Comment string for the new user (if applicable).")
        parser.add_argument('--groups',
                            help="Optional list of groups (roles) to which the new user "
                            "should be assigned.")

    def run(self, args):
        params = dict(args._get_kwargs())

        mkuser = None
        if args.system == 'rattail':
            mkuser = mkuser_rattail
        elif args.system == 'windows':
            mkuser = mkuser_windows

        if mkuser:
            if mkuser(self.config, params):
                self.stdout.write(f"created new user in '{args.system}' system: {args.username}\n")
        else:
            self.stderr.write(f"don't know how to make user for '{args.system}' system\n")
            sys.exit(1)


def mkuser_rattail(config, params):
    from sqlalchemy import orm
    from rattail.db import auth

    app = config.get_app()
    session = app.make_session()
    model = app.model
    if session.query(model.User).filter_by(username=params['username']).count():
        session.close()
        return user_exists(params)

    roles = []
    if params['groups']:
        for name in parse_list(params['groups']):
            try:
                role = session.query(model.Role)\
                              .filter(model.Role.name == name)\
                              .one()
            except orm.exc.NoResultFound:
                sys.stderr.write("Role not found: {}\n".format(name))
                session.close()
                sys.exit(4)
            else:
                roles.append(role)

    user = model.User(username=params['username'])
    if not params['no_password']:
        auth.set_user_password(user, obtain_password(params))

    if params['admin']:
        user.roles.append(auth.administrator_role(session))
    for role in roles:
        user.roles.append(role)

    if params['full_name']:
        kwargs = {'display_name': params['full_name']}
        words = params['full_name'].split()
        if len(words) == 2:
            kwargs.update({'first_name': words[0], 'last_name': words[1]})
        user.person = model.Person(**kwargs)

    session.add(user)
    session.commit()
    session.close()
    return True


def mkuser_windows(config, params, args):
    if sys.platform != 'win32':
        sys.stderr.write("sorry, only win32 platform is supported\n")
        sys.exit(1)

    from rattail.win32 import users
    from rattail.win32 import require_elevation

    if params['no_password']:
        sys.stderr.write("sorry, a password is required when making a 'win32' user\n")
        sys.exit(1)

    require_elevation()

    if users.user_exists(params['username']):
        return user_exists(params)

    return users.create_user(params['username'], obtain_password(params),
                             full_name=params['full_name'], comment=params['comment'])


def obtain_password(params):
    if params['password']:
        return params['password']
    try:
        password = None
        while not password:
            password = getpass(str("enter password for new user: ")).strip()
    except KeyboardInterrupt:
        sys.stderr.write("\noperation canceled by user\n")
        sys.exit(2)
    return password


def user_exists(params):
    sys.stdout.write(f"user already exists in '{params['system']}' system: {params['username']}\n")
    sys.exit(1)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Commands to manage settings
"""

import logging
import sys

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


log = logging.getLogger(__name__)


@rattail_typer.command()
def config_setting(
        ctx: typer.Context,
        section: Annotated[
            str,
            typer.Option(help="Section name for the config setting.")] = None,
        option: Annotated[
            str,
            typer.Option(help="Option name for the config setting.")] = None,
        name: Annotated[
            str,
            typer.Option(help="Name of the config setting to get.  "
                         "This may be used instead of --section and --option.")] = None,
        usedb: Annotated[
            bool,
            typer.Option('--usedb',
                         help="Look for values in the DB (settings table).")] = False,
        no_usedb: Annotated[
            bool,
            typer.Option('--no-usedb',
                         help="Do not look for values in the DB (settings table).")] = False,
        preferdb: Annotated[
            bool,
            typer.Option('--preferdb',
                         help="Prefer values from DB over those from config file.")] = False,
        no_preferdb: Annotated[
            bool,
            typer.Option('--no-preferdb',
                         help="Prefer values from config file over those from DB.")] = False,
):
    """
    Get a value from config file and/or settings table
    """
    config = ctx.parent.rattail_config

    # nb. we still may be using legacy config object, so must
    # convert name to (section, option) if applicable
    if section and option:
        section = section
        option = option
    elif name:
        section, option = legacy_split_setting(name)

    usedb = None
    if usedb:
        usedb = True
    elif no_usedb:
        usedb = False

    preferdb = None
    if preferdb:
        preferdb = True
    elif no_preferdb:
        preferdb = False

    value = config.get(section, option, usedb=usedb, preferdb=preferdb)
    if value is not None:
        sys.stdout.write(f"{value}\n")


@rattail_typer.command()
def setting_get(
        ctx: typer.Context,
        name: Annotated[
            str,
            typer.Argument(help="Name of the setting to retrieve.")] = ...,
):
    """
    Get a setting value from the DB
    """
    config = ctx.parent.rattail_config
    app = config.get_app()
    session = app.make_session()
    value = app.get_setting(session, name)
    session.commit()
    session.close()
    sys.stdout.write(value or '')


@rattail_typer.command()
def setting_put(
        ctx: typer.Context,
        name: Annotated[
            str,
            typer.Argument(help="Name of the setting to save.")] = ...,
        value: Annotated[
            str,
            typer.Argument(help="String value for the setting.")] = ...,
):
    """
    Add or update a setting in the DB
    """
    config = ctx.parent.rattail_config
    app = config.get_app()
    session = app.make_session()
    app.save_setting(session, name, value)
    session.commit()
    session.close()


class ConfigSetting(Subcommand):
    """
    Get a value from config file and/or settings table
    """
    name = 'config-setting'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--section', help="Section name for the config setting.")
        parser.add_argument('--option', help="Option name for the config setting.")
        parser.add_argument('--name', help="Name of the config setting to get.  "
                            "This may be used instead of --section and --option.")

        parser.add_argument('--usedb', action='store_true',
                            help="Look for values in the DB (settings table).")
        parser.add_argument('--no-usedb', action='store_true',
                            help="Do not look for values in the DB (settings table).")

        parser.add_argument('--preferdb', action='store_true',
                            help="Prefer values from DB over those from config file.")
        parser.add_argument('--no-preferdb', action='store_true',
                            help="Prefer values from config file over those from DB.")

    def run(self, args):
        """
        Retrieves a config setting via simple call to
        :meth:`~rattail.config.RattailConfig.get()` and prints
        the value to :attr:`~rattail.commands.core.Command.stdout`.
        """
        # nb. we still may be using legacy config object, so must
        # convert name to (section, option) if applicable
        if args.section and args.option:
            section = args.section
            option = args.option
        elif args.name:
            section, option = legacy_split_setting(args.name)

        usedb = None
        if args.usedb:
            usedb = True
        elif args.no_usedb:
            usedb = False

        preferdb = None
        if args.preferdb:
            preferdb = True
        elif args.no_preferdb:
            preferdb = False

        value = self.config.get(section, option, usedb=usedb, preferdb=preferdb)
        if value is not None:
            self.stdout.write(f"{value}\n")


def legacy_split_setting(name):
    """
    Split a new-style setting ``name`` into a legacy 2-tuple of
    ``(section, option)``.
    """
    parts = name.split('.')
    if len(parts) > 2:
        log.debug("ambiguous legacy split for setting name: %s", name)
    return parts[0], '.'.join(parts[1:])


class SettingGet(Subcommand):
    """
    Get a setting value from the DB
    """
    name = 'setting-get'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('name', help="Name of the setting to retrieve.")

    def run(self, args):
        session = self.make_session()
        value = self.app.get_setting(session, args.name)
        session.commit()
        session.close()
        self.stdout.write(value or '')


class SettingPut(Subcommand):
    """
    Add or update a setting in the DB
    """
    name = 'setting-put'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('name', help="Name of the setting to save.")
        parser.add_argument('value', help="String value for the setting.")

    def run(self, args):
        session = self.make_session()
        self.app.save_setting(session, args.name, args.value)
        session.commit()
        session.close()



# -*- coding: utf-8; -*-
"""add batch_purchase_row.invoice_number

Revision ID: 0d1c2362b8b7
Revises: e00841e2deb2
Create Date: 2023-07-07 17:09:03.723130

"""

# revision identifiers, used by Alembic.
revision = '0d1c2362b8b7'
down_revision = 'e00841e2deb2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # batch_purchase_row
    op.add_column('batch_purchase_row', sa.Column('invoice_number', sa.String(length=20), nullable=True))

    # purchase_item
    op.add_column('purchase_item', sa.Column('invoice_number', sa.String(length=20), nullable=True))


def downgrade():

    # purchase_item
    op.drop_column('purchase_item', 'invoice_number')

    # batch_purchase_row
    op.drop_column('batch_purchase_row', 'invoice_number')



# -*- coding: utf-8; -*-
"""grow product.true_margin

Revision ID: 492f7f75e07f
Revises: 17cd825534de
Create Date: 2023-11-15 09:28:05.215294

"""

# revision identifiers, used by Alembic.
revision = '492f7f75e07f'
down_revision = '17cd825534de'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product_volatile
    op.alter_column('product_volatile', 'true_margin',
               existing_type=sa.NUMERIC(precision=9, scale=5),
               type_=sa.Numeric(precision=12, scale=5),
               existing_nullable=True)


def downgrade():

    # product_volatile
    op.alter_column('product_volatile', 'true_margin',
               existing_type=sa.Numeric(precision=12, scale=5),
               type_=sa.NUMERIC(precision=9, scale=5),
               existing_nullable=True)



# -*- coding: utf-8; -*-
"""add Department.default_custorder_discount

Revision ID: 11dd4ffbe8c9
Revises: 492f7f75e07f
Create Date: 2023-12-22 11:38:15.504641

"""

# revision identifiers, used by Alembic.
revision = '11dd4ffbe8c9'
down_revision = '492f7f75e07f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # department
    op.add_column('department', sa.Column('default_custorder_discount', sa.Numeric(precision=5, scale=3), nullable=True))
    op.add_column('department_version', sa.Column('default_custorder_discount', sa.Numeric(precision=5, scale=3), autoincrement=False, nullable=True))


def downgrade():

    # department
    op.drop_column('department_version', 'default_custorder_discount')
    op.drop_column('department', 'default_custorder_discount')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail date-organize`` command
"""

import datetime
import os
import shutil
from pathlib import Path

import typer
from typing_extensions import Annotated

from .base import rattail_typer


@rattail_typer.command()
def date_organize(
        ctx: typer.Context,
        folder: Annotated[
            Path,
            typer.Argument(help="Path to directory containing files which are "
                           "to be organized by date.")] = ...,
):
    """
    Organize files into subfolders according to date
    """
    config = ctx.parent.rattail_config
    app = config.get_app()
    today = app.today()
    for filename in sorted(os.listdir(folder)):
        path = os.path.join(folder, filename)
        if os.path.isfile(path):
            mtime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
            if mtime.date() < today:
                datedir = mtime.strftime(os.sep.join(('%Y', '%m', '%d')))
                datedir = os.path.join(folder, datedir)
                if not os.path.exists(datedir):
                    os.makedirs(datedir)
                shutil.move(path, datedir)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail run-n-mail`` command
"""

import logging
import subprocess

import humanize
import typer
from typing_extensions import Annotated
from wuttjamaican.util import parse_list

from .base import rattail_typer, Subcommand


log = logging.getLogger(__name__)


@rattail_typer.command()
def run_n_mail(
        ctx: typer.Context,
        skip_if_empty: Annotated[
            bool,
            typer.Option('--skip-if-empty',
                         help="Skip sending the email if the command generates no output.")] = False,
        key: Annotated[
            str,
            typer.Option(help="Config key for email settings")] = 'run_n_mail',

        # TODO: these all seem like good ideas, but not needed yet?
        # parser.add_argument('--from', '-F', metavar='ADDRESS',
        #                     help="Override value of From: header")
        # parser.add_argument('--to', '-T', metavar='ADDRESS',
        #                     help="Override value of To: header (may specify more than once)")
        # parser.add_argument('--cc', metavar='ADDRESS',
        #                     help="Override value of Cc: header (may specify more than once)")
        # parser.add_argument('--bcc', metavar='ADDRESS',
        #                     help="Override value of Bcc: header (may specify more than once)")

        subject: Annotated[
            str,
            typer.Option('--subject', '-S',
                         help="Override value of Subject: header (i.e. value after prefix)")] = None,
        command: Annotated[
            str,
            typer.Argument(help="Command which should be ran, and result of which will be emailed")] = ...,
        keep_exit_code: Annotated[
            bool,
            typer.Option('--keep-exit-code',
                         help="Exit with same return code as subprocess.  If "
                         "this is not specified, `run-n-mail` will normally "
                         "exit with code 0 regardless of what happens with "
                         "the subprocess.")] = False,
):
    """
    Run a command as subprocess, and email the result/output
    """
    config = ctx.parent.rattail_config
    app = config.get_app()

    cmd = parse_list(command)
    log.info("will run command as subprocess: %s", cmd)
    run_began = app.make_utc()

    try:
        # TODO: must we allow for shell=True in some situations? (clearly not yet)
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        retcode = 0
        log.info("command completed successfully")
    except subprocess.CalledProcessError as error:
        output = error.output
        retcode = error.returncode
        log.warning("command exited with code: %s", retcode)

    output = output.decode(errors='replace')

    run_ended = app.make_utc()
    runtime = run_ended - run_began
    runtime_pretty = humanize.naturaldelta(runtime)

    if skip_if_empty and not output:
        log.info("command had no output, so will skip sending email")
        # return

    else: # send email
        kwargs = {}
        if subject:
            kwargs['subject_template'] = subject
        app.send_email(key, {
            'cmd': cmd,
            'run_began': run_began,
            'run_ended': run_ended,
            'runtime': runtime,
            'runtime_pretty': runtime_pretty,
            'retcode': retcode,
            'output': output,
        }, **kwargs)

    if retcode and keep_exit_code:
        sys.exit(retcode)


class RunAndMail(Subcommand):
    """
    Run a command as subprocess, and email the result/output
    """
    name = 'run-n-mail'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--skip-if-empty', action='store_true',
                            help="Skip sending the email if the command generates no output.")

        parser.add_argument('--key', default='run_n_mail',
                            help="Config key for email settings")
        # TODO: these all seem like good ideas, but not needed yet?
        # parser.add_argument('--from', '-F', metavar='ADDRESS',
        #                     help="Override value of From: header")
        # parser.add_argument('--to', '-T', metavar='ADDRESS',
        #                     help="Override value of To: header (may specify more than once)")
        # parser.add_argument('--cc', metavar='ADDRESS',
        #                     help="Override value of Cc: header (may specify more than once)")
        # parser.add_argument('--bcc', metavar='ADDRESS',
        #                     help="Override value of Bcc: header (may specify more than once)")
        parser.add_argument('--subject', '-S',
                            help="Override value of Subject: header (i.e. value after prefix)")
        parser.add_argument('cmd', metavar='COMMAND',
                            help="Command which should be ran, and result of which will be emailed")

        parser.add_argument('--keep-exit-code', action='store_true',
                            help="Exit with same return code as subprocess.  If "
                            "this is not specified, `run-n-mail` will normally "
                            "exit with code 0 regardless of what happens with "
                            "the subprocess.")

    def run(self, args):
        cmd = parse_list(args.cmd)
        log.info("will run command as subprocess: %s", cmd)
        run_began = self.app.make_utc()

        try:
            # TODO: must we allow for shell=True in some situations? (clearly not yet)
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            retcode = 0
            log.info("command completed successfully")
        except subprocess.CalledProcessError as error:
            output = error.output
            retcode = error.returncode
            log.warning("command exited with code: %s", retcode)

        output = output.decode(errors='replace')

        run_ended = self.app.make_utc()
        runtime = run_ended - run_began
        runtime_pretty = humanize.naturaldelta(runtime)

        if args.skip_if_empty and not output:
            log.info("command had no output, so will skip sending email")
            # return

        else: # send email
            kwargs = {}
            if args.subject:
                kwargs['subject_template'] = args.subject
            self.app.send_email(args.key, {
                'cmd': cmd,
                'run_began': run_began,
                'run_ended': run_ended,
                'runtime': runtime,
                'runtime_pretty': runtime_pretty,
                'retcode': retcode,
                'output': output,
            }, **kwargs)

        if retcode and args.keep_exit_code:
            sys.exit(retcode)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Versioning Commands
"""

import sys
import logging
from typing import List

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


log = logging.getLogger(__name__)


@rattail_typer.command()
def purge_versions(
        ctx: typer.Context,
        list_models: Annotated[
            bool,
            typer.Option('--list', '-l',
                         help="Show list of all model names, for which version tables exist.")] = False,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Purge version data for some or all tables
    """
    from rattail.db.util import finalize_session

    config = ctx.parent.rattail_config
    app = config.get_app()

    if not config.versioning_enabled():
        sys.stderr.write("Continuum versioning is not enabled, per config\n")
        sys.exit(1)

    if list_models:
        do_list_models(config)

    else:
        session = app.make_session()
        do_purge_models(config, session)
        finalize_session(session, dry_run=dry_run)


@rattail_typer.command()
def version_check(
        ctx: typer.Context,
        models: Annotated[
            List[str],
            typer.Argument(help="Which data models to check.  If you specify any, then only "
                           "data for those models will be checked.  If you do not specify "
                           "any, then all supported models will be checked.")] = None,
        list_models: Annotated[
            bool,
            typer.Option('--list', '-l',
                         help="Show list of all model names, for which version tables exist.")] = False,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                           help="Go through the full motions and allow logging etc. to "
                           "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Run consistency checks for version data tables
    """
    from rattail.db.util import finalize_session

    config = ctx.parent.rattail_config
    app = config.get_app()

    if not config.versioning_enabled():
        sys.stderr.write("Continuum versioning is not enabled, per config\n")
        sys.exit(1)

    if list_models:
        do_list_models(config)

    else:
        session = app.make_session()
        do_version_checks(config, session, models)
        finalize_session(session, dry_run=dry_run)


class VersionSubcommand(Subcommand):
    """
    Base class for subcommands which act on version data
    """

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--list', '-l', action='store_true', dest='list_models',
                            help="Show list of all model names, for which version tables exist.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")


class PurgeVersions(VersionSubcommand):
    """
    Purge version data for some or all tables
    """
    name = 'purge-versions'
    description = __doc__.strip()

    def run(self, args):
        if not self.config.versioning_enabled():
            self.stderr.write("Continuum versioning is not enabled, per config\n")
            sys.exit(1)

        if args.list_models:
            do_list_models(self.config)

        else:
            session = self.make_session()
            do_purge_models(self.config, session)
            if args.dry_run:
                session.rollback()
                log.info("dry run, so transaction was aborted")
            else:
                session.commit()
                log.info("transaction was committed")
            session.close()


class VersionCheck(VersionSubcommand):
    """
    Run consistency checks for version data tables
    """
    name = 'version-check'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('models', nargs='*', metavar='MODEL',
                            help="Which data models to check.  If you specify any, then only "
                            "data for those models will be checked.  If you do not specify "
                            "any, then all supported models will be checked.")

    def run(self, args):
        if not self.config.versioning_enabled():
            self.stderr.write("Continuum versioning is not enabled, per config\n")
            sys.exit(1)

        if args.list_models:
            do_list_models(self.config)

        else:
            session = self.make_session()
            do_version_checks(self.config, session, args.models)
            self.finalize_session(session, dry_run=args.dry_run)


def collect_models(config):
    """
    Gather and return a dict of data model classes, for which version
    tables exist.
    """
    import sqlalchemy_continuum as continuum

    app = config.get_app()
    model = app.model

    # first we collect names of "potential" model classes
    names = []
    for name in dir(model):
        obj = getattr(model, name)
        if isinstance(obj, type):
            if issubclass(obj, model.Base):
                names.append(name)

    # next we find out if each has a version class
    models = {}
    for name in sorted(names):
        cls = getattr(model, name)
        try:
            vcls = continuum.version_class(cls)
        except continuum.ClassNotVersioned:
            pass
        else:
            models[name] = cls

    return models


def do_list_models(config):
    """
    Display a list of all version tables in the DB.
    """
    models = collect_models(config)
    if models:
        for name in sorted(models):
            sys.stdout.write(f"{name}\n")
    else:
        log.warning("hm, no version classes found; is versioning enabled?")


def do_purge_models(config, session):
    """
    Purge version data for all given models.
    """
    import sqlalchemy as sa

    models = collect_models(config)
    if not models:
        log.warning("i have no models to purge!")
        return

    for name, cls in sorted(models.items()):
        purge_version_data(config, session, cls)

    sys.stdout.write("purged all data for {} version tables\n".format(len(models)))

    log.debug("will now purge data for transaction_meta table")
    session.execute(sa.text('truncate "transaction_meta"'))
    log.debug("will now purge data for transaction table")
    session.execute(sa.text('truncate "transaction"'))
    sys.stdout.write("purged all data for 2 transaction tables\n")


def purge_version_data(config, session, cls):
    """
    Purge version data for the given model class.
    """
    import sqlalchemy as sa
    import sqlalchemy_continuum as continuum

    vcls = continuum.version_class(cls)
    vtable = vcls.__table__

    log.debug("will now purge data for version table: %s", vtable.name)
    session.execute(sa.text('truncate "{}"'.format(vtable.name)))
    sys.stdout.write("purged data for: {}\n".format(vtable.name))


def do_version_checks(config, session, models):
    """
    Run version data checks for all given models.
    """
    requested = models
    all_models = collect_models(config)
    if requested:
        models = dict([(k, v)
                       for k, v in all_models.items()
                       if k in requested])
    else:
        models = all_models
    if not models:
        log.warning("i have no models to check!")
        return

    for name, cls in sorted(models.items()):
        check_versions(config, session, cls)

    log.info("checked version data for %s models", len(models))


def check_versions(config, session, cls, progress=None):
    """
    Check version data for the given model class.
    """
    import sqlalchemy_continuum as continuum

    app = config.get_app()
    model_name = cls.__name__
    log.debug("will now check version data for model: %s", model_name)

    vcls = continuum.version_class(cls)
    versions = session.query(vcls)\
                      .order_by(vcls.transaction_id)\
                      .all()
    versions_by_uuid = {}
    result = app.make_object(problems=0)

    def organize(version, i):
        versions_by_uuid.setdefault(version.uuid, []).append(version)

    app.progress_loop(organize, versions, progress,
                      message=f"Organizing version data for {model_name}")

    def check(uuid, i):
        versions = versions_by_uuid[uuid]

        # sanity check the sequence of operations
        lastop = None
        for version in versions:
            if lastop is None:
                if version.operation_type != continuum.Operation.INSERT:
                    log.warning("first version should be INSERT for %s %s",
                                model_name, uuid)
                    result.problems += 1
            elif version.operation_type == continuum.Operation.INSERT:
                if lastop == continuum.Operation.UPDATE:
                    log.warning("INSERT following UPDATE for %s %s",
                                model_name, uuid)
                    result.problems += 1
            elif version.operation_type == continuum.Operation.UPDATE:
                if lastop == continuum.Operation.DELETE:
                    log.warning("UPDATE following DELETE for %s %s",
                                model_name, uuid)
                    result.problems += 1
            lastop = version.operation_type

    app.progress_loop(check, list(versions_by_uuid.keys()), progress,
                      message=f"Checking version data for {model_name}")

    log.info("found %s problems for model: %s", result.problems, model_name)




# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Monitoring Library

This contains misc. common/shared logic for use with multiple types of
monitors, e.g. datasync, filemon etc.
"""

from __future__ import unicode_literals, absolute_import

import os
import logging
import subprocess

from rattail.config import parse_bool, parse_list


log = logging.getLogger(__name__)


class MonitorAction(object):
    """
    Base class for monitor actions.  Note that not all actions are
    class-based, but the ones which are should probably inherit from
    this class.
    """

    def __init__(self, config):
        self.config = config
        self.app = config.get_app()
        self.enum = config.get_enum()
        self.model = config.get_model()

    def __call__(self, *args, **kwargs):
        """
        This method must be implemented in the subclass; it defines
        what the action actually *does*.  The monitor daemon will
        invoke this method for all new items which are discovered.
        """
        raise NotImplementedError


class CommandAction(MonitorAction):
    """
    Simple action which can execute an arbitrary command, as a
    subprocess.  This action is meant to be invoked with a particular
    file path, which is to be acted upon.
    """

    def __init__(self, config, cmd):
        super(CommandAction, self).__init__(config)
        self.cmd = cmd

    def __call__(self, path, **kwargs):
        """
        Run the command for the given file path.
        """
        filename = os.path.basename(path)
        shell = parse_bool(kwargs.pop('shell', False))

        if shell:
            # TODO: probably shoudn't use format() b/c who knows what is in
            # that command line, that might trigger errors
            cmd = self.cmd.format(path=path, filename=filename)

        else:
            cmd = []
            for term in parse_list(self.cmd):
                term = term.replace('{path}', path)
                term = term.replace('{filename}', filename)
                cmd.append(term)

        log.debug("final command to run is: %s", cmd)
        subprocess.check_call(cmd, shell=shell)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch-related commands
"""

import datetime
import inspect
import json
import logging
import os
import sys
from pathlib import Path

import makefun
import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand, date_argument, dict_argument
from .typer import typer_get_runas_user
from rattail.progress import SocketProgress
from rattail.util import simple_error


log = logging.getLogger(__name__)


def batch_action_command_template(
        batch_type: Annotated[
            str,
            typer.Argument(help="Type of batch to be dealt with, e.g. 'vendor_catalog'")] = ...,
        batch_uuid: Annotated[
            str,
            typer.Argument(help="UUID of the batch to be processed.")] = ...,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Stub function which provides a common param signature; used with
    :func:`batch_action_command`.
    """


def batch_action_command(fn):
    """
    Decorator for batch handler commands.  Adds common params based on
    :func:`batch_action_command_template`.
    """
    original_sig = inspect.signature(fn)
    reference_sig = inspect.signature(batch_action_command_template)

    params = list(original_sig.parameters.values())
    for i, param in enumerate(reference_sig.parameters.values()):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)


@rattail_typer.command()
def make_batch(
        ctx: typer.Context,
        batch_type: Annotated[
            str,
            typer.Argument(help="Type of batch to be created, e.g. 'vendor_catalog'")] = ...,
        input_file: Annotated[
            Path,
            typer.Option(help="Path to single input file, to be used as data "
                         "source for the new batch.  (File format will vary "
                         "depending on batch type.)")] = ...,
        create_kwargs: Annotated[
            str,
            typer.Option('--kwargs',
                         parser=json.loads,
                         help="Optional JSON-encoded string containing extra "
                         "keyword arguments to be passed to the handler's batch "
                         "creation logic.")] = None,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Make a new batch, from a data file
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_make_batch, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def auto_receive(
        ctx: typer.Context,
        **kwargs
):
    """
    Auto-receive all items in a receiving batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_auto_receive, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def execute_batch(
        ctx: typer.Context,
        execute_kwargs: Annotated[
            str,
            typer.Option('--kwargs',
                         parser=json.loads,
                         help="Optional JSON-encoded string containing extra "
                         "keyword arguments to be passed to the handler's batch "
                         "execution function.")] = None,
        **kwargs
):
    """
    Execute a batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_execute_batch, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def populate_batch(
        ctx: typer.Context,
        **kwargs
):
    """
    Populate initial data for a batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_populate_batch, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def refresh_batch(
        ctx: typer.Context,
        **kwargs
):
    """
    Refresh data for a batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_refresh_batch, user, params, progress=progress)


@rattail_typer.command()
def purge_batches(
        ctx: typer.Context,
        list_types: Annotated[
            bool,
            typer.Option('--list-types',
                         help="If set, list available batch types instead of trying "
                         "to purge anything.")] = False,
        batch_type: Annotated[
            str,
            typer.Option(help="Type of batch to be purged, e.g. 'vendor_catalog'")] = ...,
        before: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Purge all batches executed prior to this date.  If not "
                         "specified, will use --before-days to calculate instead.")] = None,
        before_days: Annotated[
            int,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Number of days before the current date, to be used "
                         "as the cutoff date if --before is not specified.  Default "
                         "is 90 days before current date.")] = 90,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Purge old batches from the database
    """
    from rattail.batch.handlers import get_batch_types
    from rattail.db.util import finalize_session

    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    app = config.get_app()

    if list_types:
        keys = get_batch_types(config)
        for key in keys:
            sys.stdout.write(f"{key}\n")
        return

    handler = app.get_batch_handler(batch_type)
    session = app.make_session()

    kwargs = {'dry_run': dry_run,
              'progress': progress}
    if before:
        kwargs['before'] = app.localtime(before)
    else:
        kwargs['before_days'] = before_days

    handler.purge_batches(session, **kwargs)
    finalize_session(session, dry_run=dry_run)


class BatchHandlerCommand(Subcommand):
    """
    Base class for commands which invoke a batch handler.

    Note that if such a command fails to run its action, the error
    will be written to STDOUT as a simple message string.  If the full
    traceback is needed, logs must be consulted.  This is done for
    sake of letting the caller capture output for display to the user
    when something goes wrong.
    """

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--batch-type', metavar='KEY',
                            help="Type of batch to be dealt with, e.g. 'vendor_catalog'")
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        user = self.get_runas_user()
        params = dict(args._get_kwargs())
        do_run(self.config, args.batch_type, self.do_stuff, user, params,
               progress=self.progress)

    def do_stuff(self, params, handler, session, user, progress=None):
        raise NotImplementedError


class MakeBatch(BatchHandlerCommand):
    """
    Make a new batch, from a data file
    """
    name = 'make-batch'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--input-file', metavar='PATH', required=True,
                            help="Path to single input file, to be used as data "
                            "source for the new batch.  (File format will vary "
                            "depending on batch type.)")

        parser.add_argument('--kwargs', dest='create_kwargs', type=dict_argument, default={},
                            help="Optional JSON-encoded string containing extra "
                            "keyword arguments to be passed to the handler's batch "
                            "creation logic.")

    def do_stuff(self, params, handler, session, user, progress=None):
        """
        This will create a new batch of the specified type, then populate it
        with the given data file.
        """
        return do_make_batch(params, handler, session, user, progress=progress)


class BatchAction(BatchHandlerCommand):
    """
    Base class for commands which invoke a handler to act on a single batch.
    """

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('batch_uuid',
                            help="UUID of the batch to be populated.")

    def do_stuff(self, params, handler, session, user, progress=None):
        """
        This will invoke some action on the batch.
        """
        batch = session.get(handler.batch_model_class, params['batch_uuid'])
        if not batch:
            raise RuntimeError("Batch of type '{}' not found: {}".format(params['batch_type'], params['batch_uuid']))

        return self.action(params, handler, batch, user, progress=progress)


class AutoReceiveBatch(BatchAction):
    """
    Auto-receive all items in a receiving batch
    """
    name = 'auto-receive'
    description = __doc__.strip()

    def action(self, params, handler, batch, user, progress=None):
        return handler.auto_receive_all_items(batch, progress=progress)


class PopulateBatch(BatchAction):
    """
    Populate initial data for a batch
    """
    name = 'populate-batch'
    description = __doc__.strip()

    def action(self, params, handler, batch, user, progress=None):
        return handler.do_populate(batch, user, progress=progress)


class RefreshBatch(BatchAction):
    """
    Refresh data for a batch
    """
    name = 'refresh-batch'
    description = __doc__.strip()

    def action(self, params, handler, batch, user, progress=None):
        return handler.do_refresh(batch, user, progress=progress)


class ExecuteBatch(BatchAction):
    """
    Execute a batch
    """
    name = 'execute-batch'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--kwargs', dest='execute_kwargs', type=dict_argument, default={},
                            help="Optional JSON-encoded string containing extra "
                            "keyword arguments to be passed to the handler's batch "
                            "execution function.")

    def action(self, params, handler, batch, user, progress=None):
        return handler.do_execute(batch, user, progress=progress, **params['execute_kwargs'])


class PurgeBatches(BatchHandlerCommand):
    """
    Purge old batches from the database
    """
    name = 'purge-batches'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--before', type=date_argument, metavar='DATE',
                            help="Purge all batches executed prior to this date.  If not "
                            "specified, will use --before-days to calculate instead.")

        parser.add_argument('--before-days', type=int, default=90, metavar='DAYS',
                            help="Number of days before the current date, to be used "
                            "as the cutoff date if --before is not specified.  Default "
                            "is 90 days before current date.")

        parser.add_argument('--list-types', action='store_true',
                            help="If set, list available batch types instead of trying "
                            "to purge anything.")


    def run(self, args):
        if args.list_types:
            self.list_types()
            return

        handler = self.app.get_batch_handler(args.batch_type)
        session = self.make_session()

        kwargs = {
            'dry_run': args.dry_run,
            'progress': self.progress,
        }
        if args.before:
            before = datetime.datetime.combine(args.before, datetime.time(0))
            before = self.app.localtime(before)
            kwargs['before'] = before
        else:
            kwargs['before_days'] = args.before_days

        purged = handler.purge_batches(session, **kwargs)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

    def list_types(self):
        from rattail.batch.handlers import get_batch_types

        keys = get_batch_types(self.config)
        for key in keys:
            self.stdout.write("{}\n".format(key))


def do_run(config, batch_type, action_func, user, params, progress=None):
    from rattail.db.util import finalize_session

    app = config.get_app()
    handler = app.get_batch_handler(batch_type)
    session = app.make_session()
    model = app.model
    user = session.get(model.User, user.uuid)

    try:
        success = action_func(params, handler, session, user, progress=progress)
    except Exception as error:
        log.warning("handler action failed", exc_info=True)
        # nb. only write "simple" error string to stdout, in case
        # caller (e.g. web app) is capturing that for display to
        # user.  admin can consult logs if more info is needed.
        # we do *not* write to stderr, because logging probably
        # is already doing that; we want to avoid mixing them.
        # TODO: this admittedly seems brittle, maybe caller should
        # pass command line flags to control this?
        sys.stdout.write(simple_error(error))
        session.rollback()
        session.close()
        sys.exit(42)

    finalize_session(session, dry_run=params['dry_run'], success=success)


def do_make_batch(config, params, handler, session, user, progress=None):
    """
    This will create a new batch of the specified type, then populate it
    with the given data file.
    """
    if not os.path.exists(params['input_file']):
        raise RuntimeError("input file path does not exist: {}".format(params['input_file']))

    app = config.get_app()
    kwargs = dict(params['create_kwargs'])
    delete_if_empty = kwargs.pop('delete_if_empty', False)
    auto_execute_allowed = kwargs.pop('auto_execute_allowed', True)

    batch = handler.make_batch(session, created_by=user, **kwargs)
    handler.set_input_file(batch, params['input_file'])
    handler.do_populate(batch, user)

    if delete_if_empty:
        session.flush()
        if not batch.data_rows:
            log.debug("auto-deleting empty '%s' batch: %s", handler.batch_key, batch)
            handler.do_delete(batch, dry_run=params['dry_run'],
                              progress=progress)
            batch = None

    if batch and auto_execute_allowed and handler.auto_executable(batch):
        handler.execute(batch, user=user)
        batch.executed = app.make_utc()
        batch.executed_by = user

    return True


def do_auto_receive(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.auto_receive_all_items(batch, progress=progress)


def do_execute_batch(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.do_execute(batch, user, progress=progress, **params['execute_kwargs'])


def do_populate_batch(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.do_populate(batch, user, progress=progress)


def do_refresh_batch(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.do_refresh(batch, user, progress=progress)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail make-uuid`` command
"""

import sys

import typer

from .base import rattail_typer, Subcommand


@rattail_typer.command()
def make_uuid(
        ctx: typer.Context,
):
    """
    Generate a new UUID
    """
    config = ctx.parent.rattail_config
    app = config.get_app()
    sys.stdout.write("{}\n".format(app.make_uuid()))


class MakeUUID(Subcommand):
    """
    Generate a new UUID
    """
    name = 'make-uuid'
    description = __doc__.strip()

    def run(self, args):
        from rattail.core import get_uuid

        self.stdout.write("{}\n".format(get_uuid()))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Commands relating to Postfix
"""

import re
import sys
import subprocess
import logging
from pathlib import Path
from typing import List

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


log = logging.getLogger(__name__)


@rattail_typer.command()
def postfix_summary(
        ctx: typer.Context,
        email: Annotated[
            bool,
            typer.Option('--email',
                         help="Email the summary instead of displaying it.")] = False,
        problems_only: Annotated[
            bool,
            typer.Option('--problems-only',
                         help="Only show summary if it contains problems.")] = False,
        yesterday: Annotated[
            bool,
            typer.Option('--yesterday',
                         help="Show summary for previous day only.")] = False,
        logfile: Annotated[
            List[Path],
            # nb. we say the file(s) must exist, but not necessariy be
            # readable by current user, since ultimately this runs
            # `sudo pflogsumm` so read access is not a problem
            typer.Option(exists=True,
                         readable=False,
                         help="Path(s) to mail log file(s)")] = ['/var/log/mail.log'],
):
    """
    Generate (and maybe email) a Postfix log summary
    """
    config = ctx.parent.rattail_config
    do_summary(config, ctx.params)


class PostfixSummary(Subcommand):
    """
    Generate (and maybe email) a Postfix log summary
    """
    name = 'postfix-summary'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--email', action='store_true', default=False,
                            help="Email the summary instead of displaying it.")

        parser.add_argument('--problems-only', action='store_true',
                            help="Only show summary if it contains problems.")

        parser.add_argument('--yesterday', action='store_true',
                            help="Show summary for previous day only.")

        parser.add_argument('logfile', nargs='*', default=['/var/log/mail.log'],
                            help="Path(s) to mail log file(s)")

    def run(self, args):
        params = dict(args._get_kwargs())
        do_dummary(self.config, params)


def do_summary(config, params):
    app = config.get_app()

    # TODO: any way to get around using 'sudo' here?
    cmd = ['sudo', 'pflogsumm', '--problems-first']
    if params['yesterday']:
        cmd.extend(['-d', 'yesterday'])
    cmd.extend(params['logfile'])

    # run cmd
    log.debug("running command: %s", cmd)
    output = subprocess.check_output(cmd).decode('utf_8')

    # parse problem counts
    problems = {}
    pattern = re.compile(r'^ +(\d)   (deferred|bounced|rejected|reject warnings|held|discarded)\b')
    for line in output.split('\n'):
        match = pattern.match(line)
        if match:
            problem = match.group(2)
            if problem in problems:
                log.warning("extra line matches '%s' pattern: %s", problem, line)
            problems[problem] = int(match.group(1))

    log.info("postfix summary contained %s problems",
             sum(problems.values()))

    # bail if caller only wants to see problems and we have none
    if params['problems_only']:
        if not any(problems.values()):
            return

    # just display the summary unless we're to email it
    if params['email']:
        app.send_email('postfix_summary', {
            'output': output,
            'problems': problems,
        })
    else:
        sys.stdout.write(output)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Cleanup commands
"""

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand, date_argument


@rattail_typer.command()
def cleanup(
        ctx: typer.Context,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through motions and log actions but do not "
                         "actually commit the results.")] = False,
):
    """
    Cleanup by removing older data etc.
    """
    from rattail.db.util import finalize_session

    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    app = config.get_app()
    handler = app.get_cleanup_handler()
    session = app.make_session()
    handler.cleanup_everything(session, dry_run=dry_run, progress=progress)
    finalize_session(session, dry_run=dry_run)


class Cleanup(Subcommand):
    """
    Cleanup by removing older data etc.
    """
    name = 'cleanup'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through motions and log actions but do not "
                            "actually commit the results.")

    def run(self, args):
        self.cleanup_handler = self.app.get_cleanup_handler()
        session = self.make_session()
        self.cleanup_handler.cleanup_everything(session,
                                                dry_run=args.dry_run,
                                                progress=self.progress)
        self.finalize_session(session, dry_run=args.dry_run)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem Reports for Trainwreck
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six

from rattail.problems import ProblemReport


class MissingDatabases(ProblemReport):
    """
    Looks for missing Trainwreck databases, with the assumption that
    there should be one DB per year archived.
    """
    system_key = 'trainwreck'
    problem_key = 'missing_dbs'
    problem_title = "Missing Databases"

    def find_problems(self, **kwargs):
        trainwreck_handler = self.app.get_trainwreck_handler()
        problems = []

        # this check only makes sense when using rotation
        if trainwreck_handler.use_rotation():

            # so far we are just looking for next year's db.  this
            # "problem" is hardly urgent since we have a whole year to
            # create the db, but still need the reminder.
            next_year = six.text_type(self.app.today().year + 1)
            if next_year not in self.config.trainwreck_engines:
                problems.append(next_year)

        return problems


class CurrentNeedsPruning(ProblemReport):
    """
    Looks for "old" transactions which need to be purged from the
    "current" Trainwreck DB.
    """
    system_key = 'trainwreck'
    problem_key = 'current_needs_pruning'
    problem_title = "Current DB Needs Pruning"

    def find_problems(self, **kwargs):
        trainwreck_handler = self.app.get_trainwreck_handler()
        problems = []

        # this check only makes sense when using rotation
        if trainwreck_handler.use_rotation():

            # cutoff is 1 Jan on oldest "current" year
            current_years = trainwreck_handler.current_years()
            cutoff_year = self.app.today().year - current_years + 1
            cutoff = datetime.datetime(cutoff_year, 1, 1)
            cutoff = self.app.localtime(cutoff)

            # just count txns older than cutoff
            trainwreck_session = trainwreck_handler.make_session()
            trainwreck = trainwreck_handler.get_model()
            count = trainwreck_session.query(trainwreck.Transaction)\
                                      .filter(trainwreck.Transaction.end_time < self.app.make_utc(cutoff))\
                                      .count()
            if count:
                problems.append((count, cutoff))

        return problems



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem Report Commands
"""

import sys

from typing import List

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


@rattail_typer.command()
def problems(
        ctx: typer.Context,
        systems: Annotated[
            List[str],
            typer.Option('--system', '-s',
                         help="System for which problem checks should be performed.  You "
                         "may specify this more than once, to check multiple systems.  If "
                         "not specified, all supported systems will be checked.")] = None,
        problems: Annotated[
            List[str],
            typer.Option('--problem', '-p',
                         help="Identifies a particular problem check which should be performed.  "
                         "You may specify this more than once, to check multiple problems.  If "
                         "not specified, all supported problem reports will run.")] = None,
        list_reports: Annotated[
            bool,
            typer.Option('--list', '-l',
                         help="List available problem reports, instead of running them.  "
                         "If --system args are present, will only show those reports.")] = False,
        fix: Annotated[
            bool,
            typer.Option('--fix',
                         help="Whether to (attempt to) fix the problems, "
                         "vs. just reporting them.")] = False,
        force: Annotated[
            bool,
            typer.Option('--force',
                         help="Force running of the report(s), i.e. ignore "
                         "the \"enabled\" and scheduling flags.")] = False,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Find and/or fix, and report on problems with the data
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    app = config.get_app()
    handler = app.get_problem_report_handler(dry_run=dry_run, progress=progress)

    # try to warn user if unknown system is specified; but otherwise ignore
    supported_systems = handler.supported_systems()
    for key in (systems or []):
        if key not in supported_systems:
            sys.stderr.write(f"No problem reports exist for system: {key}\n")

    reports = handler.get_problem_reports(systems=systems, problems=problems)

    if list_reports:
        do_list_reports(config, handler, reports)
    else:
        handler.run_problem_reports(reports, fix=fix, force=force)


class Problems(Subcommand):
    """
    Find and/or fix, and report on problems with the data
    """
    name = 'problems'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--system', '-s', dest='systems', metavar='KEY', action='append',
                            help="System for which problem checks should be performed.  You "
                            "may specify this more than once, to check multiple systems.  If "
                            "not specified, all supported systems will be checked.")

        parser.add_argument('--problem', '-p', dest='problems', metavar='KEY', action='append',
                            help="Identifies a particular problem check which should be performed.  "
                            "You may specify this more than once, to check multiple problems.  If "
                            "not specified, all supported problem reports will run.")

        parser.add_argument('--list', '-l', action='store_true', dest='list_reports',
                            help="List available problem reports, instead of running them.  "
                            "If --system args are present, will only show those reports.")

        parser.add_argument('--fix', '-F', action='store_true',
                            help="Whether to (attempt to) fix the problems, vs. just reporting them.")

        parser.add_argument('--force', action='store_true',
                            help="Force running of the report(s), i.e. ignore "
                            "the \"enabled\" and scheduling flags.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        handler = self.app.get_problem_report_handler(dry_run=args.dry_run,
                                                      progress=self.progress)

        # try to warn user if unknown system is specified; but otherwise ignore
        supported_systems = handler.supported_systems()
        for key in (args.systems or []):
            if key not in supported_systems:
                self.stderr.write("No problem reports exist for system: {}\n".format(key))

        reports = handler.get_problem_reports(systems=args.systems,
                                              problems=args.problems)

        if args.list_reports:
            do_list_reports(self.config, handler, reports)
        else:
            handler.run_problem_reports(reports, fix=args.fix, force=args.force)


def do_list_reports(config, handler, reports):
    """
    List all relevant problem reports.
    """
    organized = handler.organize_problem_reports(reports)
    for system in sorted(organized):

        sys.stdout.write("\n{}\n".format(system))
        sys.stdout.write("-------------------------\n")
        for problem in sorted(organized[system]):
            sys.stdout.write("{}\n".format(problem))

    sys.stdout.write("\n")
